/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate model Data.

  Copyright (C) 2012-2012 Uwe Schulzweida, Uwe.Schulzweida@zmaw.de
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#ifndef _CONST_H
#define _CONST_H

#define  MIN_PS      20000.
#define  MAX_PS     120000.

#define  MIN_FIS   -100000.
#define  MAX_FIS    100000.

#define  MIN_T         150.
#define  MAX_T         400.

#define  MIN_Q         0.0
#define  MAX_Q         0.1

#endif  /* _CONST_H */
