complete -W "
-- \
-- \
-L \
-M \
-O \
-P \
-Q \
-R \
-S \
-V \
-W \
-a \
-b \
-f \
-g \
-h \
-k \
-m \
-r \
-s \
-t \
-v \
-z \
abs -abs \
acos -acos \
add -add \
addc -addc \
adipot -adipot \
adisit -adisit \
aexpr -aexpr \
aexprf -aexprf \
anomaly -anomaly \
asin -asin \
atan -atan \
atan2 -atan2 \
bandpass -bandpass \
beta -beta \
boxavg -boxavg \
cat -cat \
cdiread -cdiread \
cdiwrite -cdiwrite \
change_e5lsm -change_e5lsm \
change_e5mask -change_e5mask \
change_e5slm -change_e5slm \
chcode -chcode \
chisquare -chisquare \
chlevel -chlevel \
chlevelc -chlevelc \
chlevelv -chlevelv \
chltype -chltype \
chname -chname \
chparam -chparam \
chtabnum -chtabnum \
chunit -chunit \
chvar -chvar \
cloudlayer -cloudlayer \
cmd -cmd \
com -com \
command -command \
complextorect -complextorect \
consecsum -consecsum \
consects -consects \
const -const \
conv_cmor_table -conv_cmor_table \
copy -copy \
cos -cos \
coshill -coshill \
covar0 -covar0 \
covar0r -covar0r \
dayavg -dayavg \
daycount -daycount \
daylogs -daylogs \
daymax -daymax \
daymean -daymean \
daymin -daymin \
daypctl -daypctl \
daystd -daystd \
daystd1 -daystd1 \
daysum -daysum \
dayvar -dayvar \
dayvar1 -dayvar1 \
del29feb -del29feb \
delcode -delcode \
delday -delday \
delete -delete \
delname -delname \
delparam -delparam \
deltap -deltap \
deltap_fl -deltap_fl \
delvar -delvar \
detrend -detrend \
diff -diff \
diffc -diffc \
diffn -diffn \
diffp -diffp \
diffv -diffv \
div -div \
divc -divc \
divcoslat -divcoslat \
divdpm -divdpm \
divdpy -divdpy \
dumplogo -dumplogo \
dumplogs -dumplogs \
dumpmap -dumpmap \
duplicate -duplicate \
dv2ps -dv2ps \
dv2uv -dv2uv \
dv2uvl -dv2uvl \
eca_cdd -eca_cdd \
eca_cfd -eca_cfd \
eca_csu -eca_csu \
eca_cwd -eca_cwd \
eca_cwdi -eca_cwdi \
eca_cwfi -eca_cwfi \
eca_etr -eca_etr \
eca_fd -eca_fd \
eca_gsl -eca_gsl \
eca_hd -eca_hd \
eca_hwdi -eca_hwdi \
eca_hwfi -eca_hwfi \
eca_id -eca_id \
eca_pd -eca_pd \
eca_r10mm -eca_r10mm \
eca_r1mm -eca_r1mm \
eca_r20mm -eca_r20mm \
eca_r75p -eca_r75p \
eca_r75ptot -eca_r75ptot \
eca_r90p -eca_r90p \
eca_r90ptot -eca_r90ptot \
eca_r95p -eca_r95p \
eca_r95ptot -eca_r95ptot \
eca_r99p -eca_r99p \
eca_r99ptot -eca_r99ptot \
eca_rr1 -eca_rr1 \
eca_rx1day -eca_rx1day \
eca_rx5day -eca_rx5day \
eca_sdii -eca_sdii \
eca_su -eca_su \
eca_tg10p -eca_tg10p \
eca_tg90p -eca_tg90p \
eca_tn10p -eca_tn10p \
eca_tn90p -eca_tn90p \
eca_tr -eca_tr \
eca_tx10p -eca_tx10p \
eca_tx90p -eca_tx90p \
enlarge -enlarge \
enlargegrid -enlargegrid \
ensavg -ensavg \
ensbrs -ensbrs \
enscrps -enscrps \
ensmax -ensmax \
ensmean -ensmean \
ensmin -ensmin \
enspctl -enspctl \
ensrkhist_space -ensrkhist_space \
ensrkhist_time -ensrkhist_time \
ensrkhistspace -ensrkhistspace \
ensrkhisttime -ensrkhisttime \
ensroc -ensroc \
ensstd -ensstd \
ensstd1 -ensstd1 \
enssum -enssum \
ensvar -ensvar \
ensvar1 -ensvar1 \
eof -eof \
eof3d -eof3d \
eof3dspatial -eof3dspatial \
eof3dtime -eof3dtime \
eofcoeff -eofcoeff \
eofcoeff3d -eofcoeff3d \
eofspatial -eofspatial \
eoftime -eoftime \
eq -eq \
eqc -eqc \
exp -exp \
export_e5ml -export_e5ml \
export_e5res -export_e5res \
expr -expr \
exprf -exprf \
fc2gp -fc2gp \
fc2sp -fc2sp \
fdns -fdns \
filedes -filedes \
fillmiss -fillmiss \
fillmiss2 -fillmiss2 \
fisher -fisher \
fldavg -fldavg \
fldcor -fldcor \
fldcovar -fldcovar \
fldmax -fldmax \
fldmean -fldmean \
fldmin -fldmin \
fldpctl -fldpctl \
fldrms -fldrms \
fldstd -fldstd \
fldstd1 -fldstd1 \
fldsum -fldsum \
fldvar -fldvar \
fldvar1 -fldvar1 \
for -for \
fourier -fourier \
fpressure -fpressure \
gather -gather \
ge -ge \
gec -gec \
genbic -genbic \
genbil -genbil \
gencon -gencon \
gencon2 -gencon2 \
gendis -gendis \
gengrid -gengrid \
genlaf -genlaf \
gennn -gennn \
genycon -genycon \
geopotheight -geopotheight \
ggstat -ggstat \
ggstats -ggstats \
gheight -gheight \
globavg -globavg \
gp2fc -gp2fc \
gp2sp -gp2sp \
gp2spl -gp2spl \
gradsdes -gradsdes \
gridarea -gridarea \
gridboxavg -gridboxavg \
gridboxmax -gridboxmax \
gridboxmean -gridboxmean \
gridboxmin -gridboxmin \
gridboxstd -gridboxstd \
gridboxsum -gridboxsum \
gridboxvar -gridboxvar \
griddes -griddes \
griddes2 -griddes2 \
griddx -griddx \
griddy -griddy \
gridmask -gridmask \
gridverify -gridverify \
gridweights -gridweights \
gt -gt \
gtc -gtc \
harmonic -harmonic \
highpass -highpass \
histcount -histcount \
histfreq -histfreq \
histmean -histmean \
histsum -histsum \
houravg -houravg \
hourcount -hourcount \
hourmax -hourmax \
hourmean -hourmean \
hourmin -hourmin \
hourpctl -hourpctl \
hourstd -hourstd \
hourstd1 -hourstd1 \
hoursum -hoursum \
hourvar -hourvar \
hourvar1 -hourvar1 \
hpressure -hpressure \
hurr -hurr \
ifnotthen -ifnotthen \
ifnotthenc -ifnotthenc \
ifthen -ifthen \
ifthenc -ifthenc \
ifthenelse -ifthenelse \
import_amsr -import_amsr \
import_binary -import_binary \
import_cmsaf -import_cmsaf \
import_e5ml -import_e5ml \
import_e5res -import_e5res \
import_grads -import_grads \
import_obs -import_obs \
imtocomplex -imtocomplex \
info -info \
infoc -infoc \
infon -infon \
infop -infop \
infos -infos \
infov -infov \
input -input \
inputext -inputext \
inputsrv -inputsrv \
int -int \
interpolate -interpolate \
intgrid -intgrid \
intgridbil -intgridbil \
intgridcon -intgridcon \
intgridtraj -intgridtraj \
intlevel -intlevel \
intlevel3d -intlevel3d \
intlevelx -intlevelx \
intlevelx3d -intlevelx3d \
intntime -intntime \
intpoint -intpoint \
inttime -inttime \
intyear -intyear \
invertlat -invertlat \
invertlatdata -invertlatdata \
invertlatdes -invertlatdes \
invertlev -invertlev \
invertlon -invertlon \
invertlondata -invertlondata \
invertlondes -invertlondes \
isosurface -isosurface \
le -le \
lec -lec \
lmavg -lmavg \
lmean -lmean \
lmmean -lmmean \
lmstd -lmstd \
ln -ln \
log -log \
log10 -log10 \
lowpass -lowpass \
lsmean -lsmean \
lt -lt \
ltc -ltc \
map -map \
mask -mask \
maskindexbox -maskindexbox \
masklonlatbox -masklonlatbox \
maskregion -maskregion \
mastrfu -mastrfu \
max -max \
meandiff2test -meandiff2test \
meravg -meravg \
merge -merge \
mergegrid -mergegrid \
mergetime -mergetime \
mermax -mermax \
mermean -mermean \
mermin -mermin \
merpctl -merpctl \
merstd -merstd \
mersum -mersum \
mervar -mervar \
min -min \
ml2hl -ml2hl \
ml2hl_lp -ml2hl_lp \
ml2hlx -ml2hlx \
ml2hlx_lp -ml2hlx_lp \
ml2pl -ml2pl \
ml2pl_lp -ml2pl_lp \
ml2plx -ml2plx \
ml2plx_lp -ml2plx_lp \
mod -mod \
monadd -monadd \
monavg -monavg \
moncount -moncount \
mondiv -mondiv \
monlogs -monlogs \
monmax -monmax \
monmean -monmean \
monmin -monmin \
monmul -monmul \
monpctl -monpctl \
monstd -monstd \
monstd1 -monstd1 \
monsub -monsub \
monsum -monsum \
monvar -monvar \
monvar1 -monvar1 \
mrotuv -mrotuv \
mrotuvb -mrotuvb \
mul -mul \
mulc -mulc \
mulcoslat -mulcoslat \
muldoy -muldoy \
muldpm -muldpm \
muldpy -muldpy \
ncode -ncode \
ncode -ncode \
ncopy -ncopy \
ndate -ndate \
ne -ne \
nec -nec \
nint -nint \
nlevel -nlevel \
nmltest -nmltest \
nmon -nmon \
normal -normal \
npar -npar \
ntime -ntime \
nvar -nvar \
nyear -nyear \
output -output \
outputarr -outputarr \
outputbounds -outputbounds \
outputboundscpt -outputboundscpt \
outputcenter -outputcenter \
outputcenter2 -outputcenter2 \
outputcentercpt -outputcentercpt \
outputext -outputext \
outputf -outputf \
outputfld -outputfld \
outputint -outputint \
outputkey -outputkey \
outputsrv -outputsrv \
outputtab -outputtab \
outputtri -outputtri \
outputts -outputts \
outputvector -outputvector \
outputvrml -outputvrml \
outputxyz -outputxyz \
pack -pack \
pardes -pardes \
pardup -pardup \
parmul -parmul \
partab -partab \
partab2 -partab2 \
pinfo -pinfo \
pinfov -pinfov \
pow -pow \
pressure_fl -pressure_fl \
pressure_hl -pressure_hl \
random -random \
read_cmor_table -read_cmor_table \
read_e5ml -read_e5ml \
reci -reci \
regres -regres \
remap -remap \
remapbic -remapbic \
remapbil -remapbil \
remapcon -remapcon \
remapcon1 -remapcon1 \
remapcon2 -remapcon2 \
remapdis -remapdis \
remapdis1 -remapdis1 \
remapeta -remapeta \
remapeta_s -remapeta_s \
remapeta_z -remapeta_z \
remaplaf -remaplaf \
remapnn -remapnn \
remapsum -remapsum \
remapycon -remapycon \
replace -replace \
retocomplex -retocomplex \
rhopot -rhopot \
rotuvb -rotuvb \
runavg -runavg \
runmax -runmax \
runmean -runmean \
runmin -runmin \
runpctl -runpctl \
runstd -runstd \
runstd1 -runstd1 \
runsum -runsum \
runvar -runvar \
runvar1 -runvar1 \
scalllogo -scalllogo \
scatter -scatter \
sealevelpressure -sealevelpressure \
seasavg -seasavg \
seascount -seascount \
seasmax -seasmax \
seasmean -seasmean \
seasmin -seasmin \
seaspctl -seaspctl \
seasstd -seasstd \
seassum -seassum \
seasvar -seasvar \
seinfo -seinfo \
seinfoc -seinfoc \
seinfon -seinfon \
seinfop -seinfop \
selall -selall \
selcode -selcode \
seldate -seldate \
selday -selday \
select -select \
selgrid -selgrid \
selgridname -selgridname \
selhour -selhour \
selindexbox -selindexbox \
sellevel -sellevel \
sellevidx -sellevidx \
sellonlatbox -sellonlatbox \
selltype -selltype \
selmon -selmon \
selname -selname \
seloperator -seloperator \
selparam -selparam \
selrec -selrec \
selseas -selseas \
selsmon -selsmon \
selstdname -selstdname \
seltabnum -seltabnum \
seltime -seltime \
seltimestep -seltimestep \
selvar -selvar \
selyear -selyear \
selzaxis -selzaxis \
selzaxisname -selzaxisname \
setcalendar -setcalendar \
setcindexbox -setcindexbox \
setclonlatbox -setclonlatbox \
setcode -setcode \
setctomiss -setctomiss \
setdate -setdate \
setday -setday \
setgatt -setgatt \
setgatts -setgatts \
setgrid -setgrid \
setgridarea -setgridarea \
setgridmask -setgridmask \
setgridnumber -setgridnumber \
setgridtype -setgridtype \
setgriduri -setgriduri \
sethalo -sethalo \
setlevel -setlevel \
setltype -setltype \
setmisstoc -setmisstoc \
setmissval -setmissval \
setmon -setmon \
setname -setname \
setparam -setparam \
setpartab -setpartab \
setpartabc -setpartabc \
setpartabn -setpartabn \
setpartabp -setpartabp \
setpartabv -setpartabv \
setrcaname -setrcaname \
setreftime -setreftime \
setrtoc -setrtoc \
setrtoc2 -setrtoc2 \
setrtomiss -setrtomiss \
settabnum -settabnum \
settaxis -settaxis \
settime -settime \
settunits -settunits \
setunit -setunit \
setvals -setvals \
setvar -setvar \
setvrange -setvrange \
setyear -setyear \
setzaxis -setzaxis \
shifttime -shifttime \
showcode -showcode \
showdate -showdate \
showformat -showformat \
showlevel -showlevel \
showltype -showltype \
showmon -showmon \
showname -showname \
showparam -showparam \
showstdname -showstdname \
showtime -showtime \
showtimestamp -showtimestamp \
showunit -showunit \
showvar -showvar \
showyear -showyear \
sin -sin \
sincos -sincos \
sinfo -sinfo \
sinfoc -sinfoc \
sinfon -sinfon \
sinfop -sinfop \
sinfov -sinfov \
smemlogo -smemlogo \
smooth9 -smooth9 \
snamelogo -snamelogo \
sort -sort \
sortcode -sortcode \
sortlevel -sortlevel \
sortname -sortname \
sorttaxis -sorttaxis \
sorttimestamp -sorttimestamp \
sortvar -sortvar \
sp2fc -sp2fc \
sp2gp -sp2gp \
sp2gpl -sp2gpl \
sp2sp -sp2sp \
spcut -spcut \
specinfo -specinfo \
spectrum -spectrum \
sperclogo -sperclogo \
splitcode -splitcode \
splitday -splitday \
splitgrid -splitgrid \
splithour -splithour \
splitlevel -splitlevel \
splitmon -splitmon \
splitname -splitname \
splitparam -splitparam \
splitrec -splitrec \
splitseas -splitseas \
splitsel -splitsel \
splittabnum -splittabnum \
splitvar -splitvar \
splityear -splityear \
splitzaxis -splitzaxis \
sqr -sqr \
sqrt -sqrt \
ssopar -ssopar \
stdatm -stdatm \
stimelogo -stimelogo \
strbre -strbre \
strgal -strgal \
strwin -strwin \
studentt -studentt \
sub -sub \
subc -subc \
subtrend -subtrend \
szip -szip \
tan -tan \
tee -tee \
temp -temp \
template1 -template1 \
template2 -template2 \
test -test \
test2 -test2 \
testdata -testdata \
thinout -thinout \
timavg -timavg \
timcor -timcor \
timcount -timcount \
timcovar -timcovar \
timmax -timmax \
timmean -timmean \
timmin -timmin \
timpctl -timpctl \
timselavg -timselavg \
timselmax -timselmax \
timselmean -timselmean \
timselmin -timselmin \
timselpctl -timselpctl \
timselstd -timselstd \
timselstd1 -timselstd1 \
timselsum -timselsum \
timselvar -timselvar \
timselvar1 -timselvar1 \
timsort -timsort \
timstd -timstd \
timstd1 -timstd1 \
timsum -timsum \
timvar -timvar \
timvar1 -timvar1 \
tinfo -tinfo \
topo -topo \
tpnhalo -tpnhalo \
transxy -transxy \
trend -trend \
trms -trms \
tstepcount -tstepcount \
unsetgridmask -unsetgridmask \
uv2dv -uv2dv \
uv2dvl -uv2dvl \
vardes -vardes \
vardup -vardup \
varmul -varmul \
varquot2test -varquot2test \
varrms -varrms \
vct -vct \
vct2 -vct2 \
vertavg -vertavg \
vertmax -vertmax \
vertmean -vertmean \
vertmin -vertmin \
vertstd -vertstd \
vertsum -vertsum \
vertvar -vertvar \
vertwind -vertwind \
vlist -vlist \
wct -wct \
write_e5ml -write_e5ml \
writegrid -writegrid \
writerandom -writerandom \
yarbil -yarbil \
yarcon -yarcon \
yarnn -yarnn \
ydayadd -ydayadd \
ydayavg -ydayavg \
ydaydiv -ydaydiv \
ydaymax -ydaymax \
ydaymean -ydaymean \
ydaymin -ydaymin \
ydaymul -ydaymul \
ydaypctl -ydaypctl \
ydaystd -ydaystd \
ydaystd1 -ydaystd1 \
ydaysub -ydaysub \
ydaysum -ydaysum \
ydayvar -ydayvar \
ydayvar1 -ydayvar1 \
ydrunavg -ydrunavg \
ydrunmax -ydrunmax \
ydrunmean -ydrunmean \
ydrunmin -ydrunmin \
ydrunpctl -ydrunpctl \
ydrunstd -ydrunstd \
ydrunstd1 -ydrunstd1 \
ydrunsum -ydrunsum \
ydrunvar -ydrunvar \
ydrunvar1 -ydrunvar1 \
yearavg -yearavg \
yearcount -yearcount \
yearmax -yearmax \
yearmean -yearmean \
yearmin -yearmin \
yearmonavg -yearmonavg \
yearmonmean -yearmonmean \
yearpctl -yearpctl \
yearstd -yearstd \
yearstd1 -yearstd1 \
yearsum -yearsum \
yearvar -yearvar \
yearvar1 -yearvar1 \
yhouradd -yhouradd \
yhouravg -yhouravg \
yhourdiv -yhourdiv \
yhourmax -yhourmax \
yhourmean -yhourmean \
yhourmin -yhourmin \
yhourmul -yhourmul \
yhourstd -yhourstd \
yhourstd1 -yhourstd1 \
yhoursub -yhoursub \
yhoursum -yhoursum \
yhourvar -yhourvar \
yhourvar1 -yhourvar1 \
ymonadd -ymonadd \
ymonavg -ymonavg \
ymondiv -ymondiv \
ymonmax -ymonmax \
ymonmean -ymonmean \
ymonmin -ymonmin \
ymonmul -ymonmul \
ymonpctl -ymonpctl \
ymonstd -ymonstd \
ymonstd1 -ymonstd1 \
ymonsub -ymonsub \
ymonsum -ymonsum \
ymonvar -ymonvar \
ymonvar1 -ymonvar1 \
yseasavg -yseasavg \
yseasmax -yseasmax \
yseasmean -yseasmean \
yseasmin -yseasmin \
yseaspctl -yseaspctl \
yseasstd -yseasstd \
yseassum -yseassum \
yseasvar -yseasvar \
zaxisdes -zaxisdes \
zonavg -zonavg \
zonmax -zonmax \
zonmean -zonmean \
zonmin -zonmin \
zonpctl -zonpctl \
zonrange -zonrange \
zonstd -zonstd \
zonsum -zonsum \
zonvar -zonvar \
" -f cdo
